unit Display; {Fractal Landscapes 3.0 - Copyright  1987..1996, Jon Shemitz}

interface

uses WinTypes, WinProcs,
     SysUtils, Graphics, Forms,
     Global, Database;

const
  DrawingNow: boolean = False;
  AbortDraw:  boolean = False;

type
  EAbortedDrawing = class (Exception) end;

procedure ScreenColors;

procedure PrinterColors;

procedure DrawTriangle(       Canvas:    TCanvas;
			const A, B, C:  TVertex;
                              Plys:      word;
                              PointDn:   boolean);

procedure DrawVerticals(Canvas: TCanvas);

{$ifdef Debug}
const DebugString: string = '';
{$endif}
implementation

uses Main;

type
  Surfaces = record
	      Outline, Fill: TColor;
	      end;

const
  scrnLand:     Surfaces = (Outline: clLime;  Fill: clGreen);
  scrnWater:    Surfaces = (Outline: clBlue;  Fill: clNavy);
  scrnVertical: Surfaces = (Outline: clGray;  Fill: clSilver);

  prnLand:     Surfaces = (Outline: clBlack; Fill: clWhite);
  prnWater:    Surfaces = (Outline: clBlack; Fill: clWhite);
  prnVertical: Surfaces = (Outline: clBlack; Fill: clWhite);

var
  Land, Water, Vertical: Surfaces;

procedure ScreenColors;
begin
  Land     := scrnLand;
  Water    := scrnWater;
  Vertical := scrnVertical;
end;

procedure PrinterColors;
begin
  Land     := prnLand;
  Water    := prnWater;
  Vertical := prnVertical;
end;

function Surface(Outline, Fill: TColor): Surfaces;
begin
  Result.Outline := Outline;
  Result.Fill    := Fill;
end;

{$define Pascal} {$define Float}
{$ifdef Pascal}
  {$ifdef Float}
  type
    TFloatTriple = record X, Y, Z: double; end;

  function FloatTriple(T: TTriple): TFloatTriple;
  begin
    Result.X := T.X / UnitLength;
    Result.Y := T.Y / UnitLength;
    Result.Z := T.Z / UnitLength;
  end;

  function Project(const P: TTriple): TPixel; { 3D transform a point }
  var
    Delta_Y: double;
    Tr, V:    TFloatTriple;
  begin
    Tr := FloatTriple(P);
    V  := FloatTriple(VanishingPoint);
                                         
    Delta_Y  := Tr.Y / V.Y;
    Result.X := Round( DisplayWidth *
                       ((V.X - Tr.X) * Delta_Y + Tr.X));
    Result.Y := DisplayHeight -
                Round( DisplayHeight *
                       ((V.Z - Tr.Z) * Delta_Y + Tr.Z));
  end;
  {$else}
  function Project(const Tr: TTriple): TPixel; { 3D transform a point }
  var
    Delta_Y: integer;
  begin
    Delta_Y := MulDiv(Tr.Y, UnitLength, VanishingPoint.Y);
    Result.X := MulDiv( MulDiv( VanishingPoint.X - Tr.X,
                                Delta_Y, UnitLength) + Tr.X,
                        DisplayWidth, UnitLength);
    Result.Y := DisplayHeight -
                MulDiv( MulDiv( VanishingPoint.Z - Tr.Z,
                                Delta_Y, UnitLength) + Tr.Z,
                        DisplayHeight, UnitLength );
  end;
  {$endif}
{$else}
function Project(const Tr: TTriple): TPixel; assembler; { 3D transform a point }
asm
	les	di,[Tr]
	mov	si,word ptr UnitLength	{ Scaling factor }

	mov	ax,[TTriple ptr es:di].Y	{ Tr.Y }
        imul	si			{ Scale by LoWord(UnitLength) }
	idiv	VanishingPoint.Y	{ Scaled(depth/vanishing.depth) }
{DeltaY	equ	bx }
	mov     bx,ax			{ preserve Delta.Y }

	mov	ax,VanishingPoint.Z
	sub	ax,[TTriple ptr es:di].Z	{ Delta.Z }
	imul	bx			{ Delta.Z * Delta.Y }
	idiv	si			{ Unscale(Delta.Z * Delta.Y) }
	add	ax,[TTriple ptr es:di].Z	{ Tr.Z + Unscale(Delta.Z * Delta.Y) }
	mov	cx,[DisplayHeight]	{ We'll use it twice here ... }
	imul	cx			{ (Tr.Z+Delta.Z*Delta.Y)*Screen.Row }
	idiv	si			{ Unscale }
	sub	cx,ax			{ Px.Y }

	mov	ax,VanishingPoint.X
	sub	ax,[TTriple ptr es:di].X	{ Delta.X }
	imul	bx			{ Delta.X * Delta.Y }
	idiv	si			{ Unscale(Delta.X * Delta.Y) }
	add	ax,[TTriple ptr es:di].X	{ Tr.X + Unscale(Delta.X * Delta.Y) }
	imul	[DisplayWidth]		{ (Tr.X+Delta.X*Delta.Y)*Screen.Col }
	idiv	si			{ Px.X := Unscale(above) }

        mov     dx,cx                    {Return (X,Y) in ax:dx}
end;
{$endif}

procedure DrawPixels(const Canvas:               TCanvas;
                     const A, B, C, D:           TPixel;
                     const N:                    word;
                     const Surface:              Surfaces);
begin
  if AbortDraw then raise EAbortedDrawing.Create('');

  Canvas.Pen.Color := Surface.Outline;
  if DrawMode = dmOutline
    then if N = 3
      then Canvas.PolyLine( [A, B, C, A] )
      else Canvas.PolyLine( [A, B, C, D, A] )
    else begin
         Canvas.Brush.Color := Surface.Fill;
         if N = 3
           then Canvas.Polygon( [A, B, C] )
           else Canvas.Polygon( [A, B, C, D] )
         end;
end;

procedure CalcCrossing(var Low, High, Crossing: TTriple; SetLow: boolean);
var
  CrossOverRatio: LongInt;
begin
   CrossOverRatio := (SeaLevel - Low.Z) * UnitLength div (High.Z - Low.Z);
   { Distance of crossing point from A, as ratio of total line AB length, }
   { times UnitLength                                                     }
   Crossing := Triple( Low.X + Unscale((High.X - Low.X) * CrossOverRatio),
                       Low.Y + Unscale((High.Y - Low.Y) * CrossOverRatio),
                       SeaLevel );
   if SetLow then Low.Z := SeaLevel;
end;

procedure DrawVertical(Canvas: TCanvas; const A, B: TTriple; var pA, pB: TPixel);
var
  pC, pD: TPixel;    px: TPixel;
  tC, tD: TTriple;
begin
  tC := A; tC.Z := SeaLevel; pC := Project(tC);
  tD := B; tD.Z := SeaLevel; pD := Project(tD);
  DrawPixels(Canvas, pA, pB, pD, pC, 4, Vertical);
end;

procedure DrawVerticals(Canvas: TCanvas);
type
  Triad = record
          T: TTriple;
          V: TVertex;
          P: TPixel;
          end;
var
  Work: Triad;

  procedure Step( const Start: TVertex;
                  var Front:   Triad;
                  var StepDn:  GridCoordinate
		);
  var
    Idx: word;
    Back, Interpolate: Triad;
  begin
    Back.V := Start;
    Back.T := GetTriple(Back.V);
    if Back.T.Z > SeaLevel then Back.P := Project(Back.T);
    for Idx := 1 to EdgeLength do
      begin
      Front.V := Back.V;
      Inc(Work.V.BC);
      Dec(StepDn);
      Front.T := GetTriple(Front.V);
      if Front.T.Z > SeaLevel then Front.P := Project(Front.T);
      case (ord(Back.T.Z > SeaLevel) shl 1) + ord(Front.T.Z > SeaLevel) of
        1: begin { Back below, front above }
           CalcCrossing(Back.T, Front.T, Interpolate.T, False);
           Interpolate.P := Project(Interpolate.T);
           DrawVertical(Canvas, Interpolate.T, Front.T, Interpolate.P, Front.P);
           end;
        2: begin { Back above, front below }
           CalcCrossing(Front.T, Back.T, Interpolate.T, False);
	   Interpolate.P := Project(Interpolate.T);
           DrawVertical(Canvas, Back.T, Interpolate.T, Back.P, Interpolate.P);
           end;
        3: DrawVertical(Canvas, Back.T, Front.T, Back.P, Front.P);
                 { Both above }
        end;
      Back := Front;
      end;
  end;

begin
  Step(C, Work, Work.V.AB );
  Step(B, Work, Work.V.CA );
end;

  function InnerProduct({const} A, B: TTriple): LongInt;
  begin
    InnerProduct := IMUL(A.X, B.X) + IMUL(A.Y, B.Y) + IMUL(A.Z, B.Z) ;
    { Damn but this >should< be UnScaled ...                         }
  end;

  function Delta(A, B: TTriple): TTriple;
  begin
    Result := Triple(A.X - B.X, A.Y - B.Y, A.Z - B.Z);
  end;

function LandColor(const A, B, C: TTriple): TColor;
var
  Center, ToA, ToLight: TTriple;
  Cos, Angle:           double;
  GrayLevel:            integer;
begin
  Center := Triple( (A.X + B.X + C.X) div 3,
                    (A.Y + B.Y + C.Y) div 3,
                    (A.Z + B.Z + C.Z) div 3 );
  ToA := Delta(A, Center);
  ToLight := Delta(Center, LightSource);

  {$ifopt R-} {$define ResetR} {$endif}
  {$R+}
    try
      Cos := InnerProduct(ToA, ToLight) /
             (Sqrt({Abs(}InnerProduct(ToA, ToA){)}) *
              Sqrt({Abs(}InnerProduct(ToLight, ToLight){)}) );
      try
        Angle := ArcTan (Sqrt (1 - Sqr (Cos)) / Cos);
      except
        on Exception do Angle := Pi / 2; {ArcCos(0)}
      end;
      {$ifdef HighContrast}
      GrayLevel := 255 - Round(255 * (Abs(Angle) / (Pi / 2)));
      {$else}
      GrayLevel := 235 - Round(180 * (Abs(Angle) / (Pi / 2)));
      {$endif}
    except
      on {any} Exception do GrayLevel := 255; {division by 0 ...}
    end;
  {$ifdef ResetR} {$R-} {$undef ResetR} {$endif}

  Result := PaletteRGB(GrayLevel, GrayLevel, GrayLevel);
end;

procedure Draw3Vertices(Canvas: TCanvas; const A, B, C: TVertex; Display: boolean);
var
  Color: TColor;
  pA, pB, pC, pD, pE: TPixel;
  tA, tB, tC, tD, tE: TTriple;
  aBelow, bBelow, cBelow: boolean;
begin
  tA := GetTriple(A); tB := GetTriple(B); tC := GetTriple(C);
  {$ifdef FloatingTriangles}
  ta.z := ta.z + random(Envelope shr Plys) - random(Envelope shr Plys);
  tb.z := tb.z + random(Envelope shr Plys) - random(Envelope shr Plys);
  tc.z := tc.z + random(Envelope shr Plys) - random(Envelope shr Plys);
  {$endif}
  aBelow := tA.Z <= SeaLevel;
  bBelow := tB.Z <= SeaLevel;
  cBelow := tC.Z <= SeaLevel;
  case ord(aBelow) + ord(bBelow) + ord(cBelow) of
    0:       if Display then {All above}
               begin
               pA := Project(tA); pB := Project(tB); pC := Project(tC);
               if DrawMode = dmRender
                 then begin
                      Color := LandColor(tA, tB, tC);
                      DrawPixels( Canvas,
                                  pA, pB, pC, pC, 3, Surface(Color, Color));
                      end
                 else DrawPixels( Canvas,
                                  pA, pB, pC, pC, 3, Land);
               end;
    3:       if Display then {All below}
	       begin
               tA.Z := SeaLevel; tB.Z := SeaLevel; tC.Z := SeaLevel;
               pA := Project(tA); pB := Project(tB); pC := Project(tC);
               DrawPixels( Canvas, pA, pB, pC, pC, 3, Water);
               end;
    2:       begin {One vertex above water}
             { First ensure it's tA }
             if aBelow then
               if bBelow
                 then SwapTriples(tA, tC)
                 else SwapTriples(tA, tB);
             CalcCrossing(tB, tA, tD, True);
             CalcCrossing(tC, tA, tE, True);
             pA := Project(tA); pB := Project(tB); pC := Project(tC);
             pD := Project(tD); pE := Project(tE);
             DrawPixels( Canvas, pD, pB, pC, pE, 4, Water);
	     if Drawmode = dmRender
               then begin
                    Color := LandColor(tD, tA, tE);
                    DrawPixels( Canvas, pD, pA, pE, pE, 3, Surface(Color, Color));
		    end
               else DrawPixels( Canvas, pD, pA, pE, pE, 3, Land);
             end;
    1:       begin {One vertex below water}
             { First ensure it's tA }
	     if bBelow
               then SwapTriples(tA, tB)
               else if cBelow then SwapTriples(tA, tC);
             CalcCrossing(tA, tB, tD, False);
             CalcCrossing(tA, tC, tE, True);
             pA := Project(tA); pB := Project(tB); pC := Project(tC);
             pD := Project(tD); pE := Project(tE);
             DrawPixels( Canvas, pD, pA, pE, pE, 3, Water);
             if DrawMode = dmRender
               then begin
                    Color := LandColor(tD, tB, tC);
		    DrawPixels( Canvas,
                                pD, pB, pC, pE, 4, Surface(Color, Color));
		    end
	       else DrawPixels( Canvas, pD, pB, pC, pE, 4, Land);
	     end;
   end;
end;

procedure DrawTriangle(       Canvas:   TCanvas;
		        const A, B, C:  TVertex;
                              Plys:     word;
                              PointDn:  boolean);
var
  AB, BC, CA: TVertex;
begin
  if Plys = 1
    then Draw3Vertices(Canvas, A, B, C, (DrawMode <> dmOutline) OR PointDn)
    else
      begin
      AB := Midpoint(A, B);
      BC := Midpoint(B, C);
      CA := Midpoint(C, A);
      if Plys = 3 then FractalLandscape.DrewSomeTriangles(16);
      Dec(Plys);
      if PointDn
        then begin
             DrawTriangle(Canvas, CA, BC, C, Plys, True);
             DrawTriangle(Canvas, AB, B, BC, Plys, True);
             DrawTriangle(Canvas, BC, CA, AB, Plys, False);
             DrawTriangle(Canvas, A, AB, CA, Plys, True);
             end
        else begin
             DrawTriangle(Canvas, A, CA, AB, Plys, False);
             DrawTriangle(Canvas, BC, CA, AB, Plys, True);
             DrawTriangle(Canvas, CA, C, BC, Plys, False);
             DrawTriangle(Canvas, AB, BC, B, Plys, False);
             end;
      end;
end;

begin
  ScreenColors;
end.



